import yaml
from pathlib import Path
from dataclasses import dataclass, field
from typing import Dict, List

@dataclass
class Corrections:
    """
    A dataclass to encapsulate both simple and pattern-based corrections.
    """
    simple: Dict[str, str] = field(default_factory=dict)
    patterns: List[Dict[str, str]] = field(default_factory=list)
    
    
def load_correction_list(filepath: str | Path) -> Corrections:
    """
    Load the correction dictionary and patterns from a YAML file.

    :param filepath: Path to the YAML correction file.
    :return: Corrections object containing simple and pattern-based 
        corrections.
    """
    try:
        with open(filepath, 'r') as file:
            data = yaml.safe_load(file)

        if not isinstance(data, dict):
            raise ValueError("→ Correction file must contain a YAML dictionary.")

        # Extract simple corrections
        simple_corrections = {k: v for k, v in data.items() if k != 'patterns'}

        # Extract pattern-based corrections
        pattern_corrections = data.get('patterns', [])

        # Validate patterns
        for entry in pattern_corrections:
            if 'pattern' not in entry or 'replacement' not in entry:
                raise ValueError("→ Each pattern entry must contain 'pattern' \
and 'replacement' keys.")

        return Corrections(
            simple=simple_corrections, 
            patterns=pattern_corrections
            )

    except FileNotFoundError:
        print(f"→ Correction file not found: {filepath}")
        return Corrections()
    except yaml.YAMLError as e:
        print(f"→ Error parsing YAML file: {e}")
        return Corrections()
    except Exception as e:
        print(f"→ Unexpected error loading correction list: {e}")
        return Corrections()



# Code with additional changes from: https://github.com/Vaibhavs10/insanely-fast-whisper

import requests
import torch
import numpy as np
from torchaudio import functional as F
from pyannote.audio import Pipeline
from transformers.pipelines.audio_utils import ffmpeg_read
from rich.progress import Progress, TimeElapsedColumn, SpinnerColumn, TextColumn
import sys

from whisply import little_helper

# Code lifted from https://github.com/huggingface/speechbox/blob/main/src/speechbox/diarize.py
# and from https://github.com/m-bain/whisperX/blob/main/whisperx/diarize.py

def preprocess_inputs(inputs):
    if isinstance(inputs, str):
        if inputs.startswith("http://") or inputs.startswith("https://"):
            # We need to actually check for a real protocol, otherwise it's impossible to use a local file
            # like http_huggingface_co.png
            inputs = requests.get(inputs).content
        else:
            with open(inputs, "rb") as f:
                inputs = f.read()

    if isinstance(inputs, bytes):
        inputs = ffmpeg_read(inputs, 16000)

    if isinstance(inputs, dict):
        # Accepting `"array"` which is the key defined in `datasets` for better integration
        if not ("sampling_rate" in inputs and ("raw" in inputs or "array" in inputs)):
            raise ValueError(
                "When passing a dictionary to ASRDiarizePipeline, the dict needs to contain a "
                '"raw" key containing the numpy array representing the audio and a "sampling_rate" key, '
                "containing the sampling_rate associated with that array"
            )

        _inputs = inputs.pop("raw", None)
        if _inputs is None:
            # Remove path which will not be used from `datasets`.
            inputs.pop("path", None)
            _inputs = inputs.pop("array", None)
        in_sampling_rate = inputs.pop("sampling_rate")
        inputs = _inputs
        if in_sampling_rate != 16000:
            inputs = F.resample(
                torch.from_numpy(inputs), in_sampling_rate, 16000
            ).numpy()

    if not isinstance(inputs, np.ndarray):
        raise ValueError(f"We expect a numpy ndarray as input, got `{type(inputs)}`")
    if len(inputs.shape) != 1:
        raise ValueError(
            "We expect a single channel audio input for ASRDiarizePipeline"
        )

    # diarization model expects float32 torch tensor of shape `(channels, seq_len)`
    diarizer_inputs = torch.from_numpy(inputs).float()
    diarizer_inputs = diarizer_inputs.unsqueeze(0)

    return inputs, diarizer_inputs


def diarize_audio(diarizer_inputs, diarization_pipeline, num_speakers, min_speakers, max_speakers):
    diarization = diarization_pipeline(
        {"waveform": diarizer_inputs, "sample_rate": 16000},
        num_speakers=num_speakers,
        min_speakers=min_speakers,
        max_speakers=max_speakers,
    )

    segments = []
    for segment, track, label in diarization.itertracks(yield_label=True):
        segments.append(
            {
                "segment": {"start": segment.start, "end": segment.end},
                "track": track,
                "label": label,
            }
        )

    # diarizer output may contain consecutive segments from the same speaker (e.g. {(0 -> 1, speaker_1), (1 -> 1.5, speaker_1), ...})
    # we combine these segments to give overall timestamps for each speaker's turn (e.g. {(0 -> 1.5, speaker_1), ...})
    new_segments = []
    prev_segment = cur_segment = segments[0]

    for i in range(1, len(segments)):
        cur_segment = segments[i]

        # check if we have changed speaker ("label")
        if cur_segment["label"] != prev_segment["label"] and i < len(segments):
            # add the start/end times for the super-segment to the new list
            new_segments.append(
                {
                    "segment": {
                        "start": prev_segment["segment"]["start"],
                        "end": cur_segment["segment"]["start"],
                    },
                    "speaker": prev_segment["label"],
                }
            )
            prev_segment = segments[i]

    # add the last segment(s) if there was no speaker change
    new_segments.append(
        {
            "segment": {
                "start": prev_segment["segment"]["start"],
                "end": cur_segment["segment"]["end"],
            },
            "speaker": prev_segment["label"],
        }
    )

    return new_segments


def post_process_segments_and_transcripts(new_segments, transcript, group_by_speaker) -> list:
    # get the end timestamps for each chunk from the ASR output
    end_timestamps = np.array(
        [chunk["timestamp"][-1] if chunk["timestamp"][-1] is not None else sys.float_info.max for chunk in transcript])
    segmented_preds = []

    # align the diarizer timestamps and the ASR timestamps
    for segment in new_segments:
        # get the diarizer end timestamp
        end_time = segment["segment"]["end"]
        # find the ASR end timestamp that is closest to the diarizer's end timestamp and cut the transcript to here
        upto_idx = np.argmin(np.abs(end_timestamps - end_time))

        if group_by_speaker:
            segmented_preds.append(
                {
                    "speaker": segment["speaker"],
                    "text": "".join(
                        [chunk["text"] for chunk in transcript[: upto_idx + 1]]
                    ),
                    "timestamp": (
                        transcript[0]["timestamp"][0],
                        transcript[upto_idx]["timestamp"][1],
                    ),
                }
            )
        else:
            for i in range(upto_idx + 1):
                segmented_preds.append({"speaker": segment["speaker"], **transcript[i]})

        # crop the transcripts and timestamp lists according to the latest timestamp (for faster argmin)
        transcript = transcript[upto_idx + 1:]
        end_timestamps = end_timestamps[upto_idx + 1:]

        if len(end_timestamps) == 0:
            break 

    return segmented_preds

def diarize(outputs, **kwargs):
    diarization_pipeline = Pipeline.from_pretrained(
        checkpoint_path=kwargs['diarization_model'],
        use_auth_token=kwargs['hf_token'],
    )
    diarization_pipeline.to(torch.device("mps"))

    _, diarizer_inputs = preprocess_inputs(
        inputs=kwargs['file_name']
        )

    segments = diarize_audio(
        diarizer_inputs, 
        diarization_pipeline, 
        kwargs['num_speakers'], 
        kwargs['min_speakers'], 
        kwargs['max_speakers']
    )

    return post_process_segments_and_transcripts(
        segments, outputs["chunks"], group_by_speaker=False
    )



import logging
import time
from pathlib import Path
from datetime import datetime
from functools import partial
from rich import print

from whisply import little_helper, output_utils, models
from whisply.little_helper import FilePathProcessor
from whisply.post_correction import Corrections

# Set logging configuration
logging.basicConfig(
    filename=f"log_whisply_{datetime.now().strftime('%Y-%m-%d')}.log", 
    level=logging.INFO, format='%(asctime)s %(levelname)s [%(funcName)s]: %(message)s'
    )

class TranscriptionHandler:
    """
    Handles transcription and diarization of audio/video files using various 
    Whisper-based models.

    This class leverages different implementations of OpenAI's Whisper models 
    (whisperX, insanely-fast-whisper, faster-whisper) to transcribe audio and 
    video files. It supports features like language detection, speaker 
    diarization, translation, subtitle generation, and exporting transcriptions
    in multiple formats. It is capable of processing single files, directories, 
    URLs, and lists of files, providing flexibility for diverse transcription 
    needs.

    Args:
        base_dir (str, optional): Directory to store transcription outputs. 
            Defaults to './transcriptions'.
        model (str, optional): Whisper model variant to use (e.g., 'large-v2'). 
            Defaults to 'large-v3-turbo'.
        device (str, optional): Compute device ('cpu', 'cuda', etc.). 
            Defaults to 'cpu'.
        file_language (str, optional): Language of the input audio. 
            If not provided, language detection is performed.
        annotate (bool, optional): Enable speaker diarization. 
            Defaults to False.
        hf_token (str, optional): Hugging Face token for accessing restricted  
            models or features.
        subtitle (bool, optional): Generate subtitles with word-level timestamps. 
            Defaults to False.
        sub_length (int, optional): Maximum number of words per subtitle chunk. 
            Required if subtitle is True.
        translate (bool, optional): Translate transcription to English if the 
            original language is different. Defaults to False.
        verbose (bool, optional): Enable detailed logging and output. 
            Defaults to False.
        export_formats (str or list, optional): Formats to export transcriptions 
            (e.g., 'json', 'srt'). Defaults to 'all'.

    Attributes:
        base_dir (Path): Directory for storing transcriptions.
        device (str): Compute device in use.
        file_language (str or None): Detected or specified language of the 
            audio.
        annotate (bool): Indicates if speaker diarization is enabled.
        translate (bool): Indicates if translation is enabled.
        subtitle (bool): Indicates if subtitle generation is enabled.
        verbose (bool): Indicates if verbose mode is active.
        export_formats (str or list): Selected formats for exporting 
            transcriptions.
        processed_files (list): List of processed file information and results.

    Methods:
        get_filepaths(filepath: str):
            Retrieves and validates file paths from various input types.
        
        detect_language(file: Path, audio_array) -> str:
            Detects the language of the given audio file.
        
        process_files(files: list):
            Processes a list of audio files for transcription and diarization.
        
        transcribe_with_whisperx(filepath: Path) -> dict:
            Transcribes an audio file using the whisperX implementation.
        
        transcribe_with_insane_whisper(filepath: Path) -> dict:
            Transcribes an audio file using the insanely-fast-whisper 
            implementation.
        
        transcribe_with_faster_whisper(filepath: Path, num_workers: int = 1) 
            -> dict:
            Transcribes an audio file using the faster-whisper implementation.
        
        adjust_word_chunk_length(result: dict) -> dict:
            Splits transcription text into chunks based on a maximum word 
            count.
        
        to_transcription_dict(insanely_annotation: list[dict]) -> dict:
            Converts speaker-annotated results into a standardized dictionary.
        
        to_whisperx(transcription_result: dict) -> dict:
            Normalizes transcription results to the whisperX format.
        
        create_text_with_speakers(transcription_dict: dict, 
            delimiter: str = '.') -> dict:
            Inserts speaker labels into the transcription text upon speaker 
            changes.
    """
    def __init__(
        self, 
        base_dir='./transcriptions', 
        model='large-v3-turbo', 
        device='cpu', 
        file_language=None, 
        annotate=False,
        num_speakers=None,
        hf_token=None, 
        subtitle=False, 
        sub_length=None, 
        translate=False, 
        verbose=False,
        del_originals=False,
        corrections=Corrections,
        export_formats='all'
    ):
        self.base_dir = little_helper.ensure_dir(Path(base_dir))
        self.file_formats = little_helper.return_valid_fileformats()
        self.device = device
        self.file_language = file_language
        self.file_language_provided = file_language is not None
        self.model = None
        self.model_provided = model
        self.annotate = annotate
        self.num_speakers = num_speakers
        self.translate = translate
        self.hf_token = hf_token
        self.subtitle = subtitle
        self.sub_length = sub_length
        self.verbose = verbose
        self.del_originals = del_originals
        self.corrections = corrections
        self.export_formats = export_formats
        self.metadata = self._collect_metadata()
        self.filepaths = []
        self.output_dir = None
        self.processed_files = []

    def _collect_metadata(self):
        return {
            'output_dir': str(self.base_dir),
            'file_language': self.file_language,
            'model': self.model_provided,
            'device': self.device,
            'annotate': self.annotate,
            'num_speakers': self.num_speakers,
            'translate': self.translate,
            'subtitle': self.subtitle,
            'sub_length': self.sub_length
            }
    
    def adjust_word_chunk_length(self, result: dict) -> dict:
        """
        Generates text chunks based on the maximum number of words.

        Parameters:
            result (dict): The nested dictionary containing segments 
            and words.
            max_number (int): The maximum number of words per chunk. 
            Default is 6.

        Returns:
            dict: A dictionary containing a list of chunks, each with 
            'text', 'timestamp', and 'words'.
        """
        # Flatten all words from all segments
        words = [
            word_info
            for segment in result.get('segments', [])
            for word_info in segment.get('words', [])
        ]

        # Split words into chunks of size max_number
        def split_into_chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i:i + n]

        chunks = []
        for word_chunk in split_into_chunks(words, self.sub_length):
            chunk_text = ' '.join(word_info['word'] for word_info in word_chunk)
            chunk_start = word_chunk[0]['start']
            chunk_end = word_chunk[-1]['end']
            chunk = {
                'timestamp': [chunk_start, chunk_end],
                'text': chunk_text,
                'words': word_chunk
            }
            chunks.append(chunk)

        result_temp = {
            'text': ' '.join(chunk['text'].strip() for chunk in chunks),
            'chunks': chunks
        }

        return result_temp

    def to_transcription_dict(self, insanely_annotation: list[dict]) -> dict:
        """
        Transform insanely-fast-whisper speaker annotation result to dict.
        """
        chunks = []
        for s in insanely_annotation:
            chunk = {
                'text': s['text'],
                'timestamp': (s['timestamp'][0], s['timestamp'][1]),
                'speaker': s['speaker']
            }
            chunks.append(chunk)
            
        result = {
            'text': ''.join([s['text'] for s in insanely_annotation]),
            'chunks': chunks
        }
        return result

    def to_whisperx(self, transcription_result: dict) -> dict:
        """
        Normalize insanely-fast-whisper transcription result to whisperX dict.
        """
        words = []
        for c in transcription_result['chunks']:
            if 'speaker' in c:
                word = {
                    'word': c['text'].strip(),
                    'start': c['timestamp'][0],
                    'end': c['timestamp'][1],
                    'speaker': c['speaker']
                }
            else:
                word = {
                    'word': c['text'].strip(),
                    'start': c['timestamp'][0],
                    'end': c['timestamp'][1]
                }
            words.append(word)

        result = {
            'segments': [
                {
                    'start': transcription_result['chunks'][0]['timestamp'][0],
                    'end': transcription_result['chunks'][-1]['timestamp'][1],
                    'text': transcription_result['text'].strip(),
                    'words': words
                }
            ]
        }
        return result
    
    def create_text_with_speakers(
        self,
        transcription_dict: dict, 
        delimiter: str = '.'
        ) -> dict:
        """
        Iterates through all chunks of each language and creates the complete 
        text with speaker labels inserted when there is a speaker change.

        Args:
            transcription_dict (dict): The dictionary containing transcription 
            data.

        Returns:
            dict: A dictionary mapping each language to its formatted text with 
            speaker labels.
        """    
        transcriptions = transcription_dict.get('transcriptions', {})
        
        for lang, lang_data in transcriptions.items():
            text = ""
            current_speaker = None
            chunks = lang_data.get('chunks', [])
            
            for chunk in chunks:
                words = chunk.get('words', [])
                
                for word_info in words:
                    speaker = word_info.get('speaker')
                    word = word_info.get('word', '')
                    start_timestamp = little_helper.format_time(
                        word_info.get('start'), 
                        delimiter
                        )
                    
                    # Insert speaker label if a speaker change is detected
                    if speaker != current_speaker:
                        text += f"\n[{start_timestamp}] [{speaker}] "
                        current_speaker = speaker
                    
                    # Append the word with a space
                    text += word + " "
            
            transcription_dict['transcriptions'][lang]['text_with_speaker_annotation'] = text.strip()
        
        return transcription_dict

    def transcribe_with_whisperx(self, filepath: Path) -> dict:
        """
        Transcribe a file with the whisperX implementation that returns word-level
        timestamps and speaker annotation:https://github.com/m-bain/whisperX 
        
        This implementation is used when a specific subtitle length (e.g. 5 words 
        per individual subtitle) is needed.
        """
        import torch
        import whisperx
        import gc
        
        def empty_cuda_cache(model):
            gc.collect()
            torch.cuda.empty_cache()
            del model
        
        def fill_missing_timestamps(segments: list) -> list:
            """
            whisperX does not provide timestamps for words containing only 
            numbers (e.g. "1.5", "2024" etc.).
            
            The function fills these missing timestamps by padding the last 
            known 'end' timestamp for a missing 'start' timestamp or by cutting 
            the next known 'start' timestamp for a missing 'end' timestamp.
            """
            padding = 0.05 # in seconds
            
            for segment in segments:
                words = segment['words']
                num_words = len(words)
                
                for i, word in enumerate(words):
                    # If the 'start' key is missing
                    if 'start' not in word:
                        if i > 0 and 'end' in words[i-1]:
                            word['start'] = round(words[i-1]['end'] + padding, 2)
                        else:
                            word['start'] = segment['start']
                    
                    # If the 'end' key is missing
                    if 'end' not in word:
                        if i < num_words - 1 and 'start' in words[i+1]:
                            word['end'] = round(words[i+1]['start'] - padding, 2)
                        elif i == num_words - 1:
                            word['end'] = round(words[i]['start'] + padding, 2)
                            segment['end'] = word['end']
                        else:
                            word['end'] = round(words[i]['start'] + padding, 2)
                    
                    # If 'score' key is missing       
                    if 'score' not in word:
                        word['score'] = 0.5
        
                    # If 'speaker' key is missing
                    if self.annotate and 'speaker' not in word:
                        speaker_assigned = False
                        
                         # Case 1: If it's the first word, look forward for the 
                         # next speaker
                        if i == 0:
                            for j in range(i + 1, num_words):
                                if 'speaker' in words[j]:
                                    word['speaker'] = words[j]['speaker']
                                    speaker_assigned = True
                                    break

                        # Case 2: If it's the last word, look backward for the 
                        # previous speaker
                        elif i == num_words - 1:
                            for j in range(i - 1, -1, -1):
                                if 'speaker' in words[j]:
                                    word['speaker'] = words[j]['speaker']
                                    speaker_assigned = True
                                    break

                        # Case 3: For other words, prefer the previous speaker; 
                        # If not found, look forward
                        if not speaker_assigned:
                            # Look backward
                            for j in range(i - 1, -1, -1):
                                if 'speaker' in words[j]:
                                    word['speaker'] = words[j]['speaker']
                                    speaker_assigned = True
                                    break

                        if not speaker_assigned:
                            # Look forward
                            for j in range(i + 1, num_words):
                                if 'speaker' in words[j]:
                                    word['speaker'] = words[j]['speaker']
                                    speaker_assigned = True
                                    break

                        if not speaker_assigned:
                            # Default speaker if none found
                            word['speaker'] = 'UNKNOWN'
                            
            return segments
        
        def whisperx_task(task: str = 'transcribe', language = None):
            """
            Define a transcription / translation task with whisperX
            """
            # Set parameters
            device = 'cuda' if self.device == 'cuda:0' else 'cpu'
            
            # Transcribe or Translate
            model = whisperx.load_model(
                whisper_arch=self.model, 
                device=device,
                compute_type='float16' if self.device == 'cuda:0' else 'int8', 
                language=self.file_language or None,
                asr_options={
                    "hotwords": None,
                    "multilingual": False
                })
            audio = whisperx.load_audio(str(filepath), sr=16000)
            result = model.transcribe(
                audio, 
                batch_size=16 if self.device == 'cuda:0' else 8, 
                task=task
                )
            model_a, metadata = whisperx.load_align_model(
                device=device, 
                language_code=language
                )
            result = whisperx.align(
                result["segments"], 
                model_a, 
                metadata, 
                audio, 
                device,
                return_char_alignments=False
                )
                
            # Empty CUDA cache
            if self.device == 'cuda:0':
                empty_cuda_cache(model_a)
                    
            return result
        
        def whisperx_annotation(transcription_result: dict) -> dict:
            # Set parameters
            device = 'cuda' if self.device == 'cuda:0' else 'cpu'
            
            diarize_model = whisperx.DiarizationPipeline(
                use_auth_token=self.hf_token, 
                device=device
                )
            diarize_segments = diarize_model(
                str(filepath),
                max_speakers=self.num_speakers
                )
            result = whisperx.assign_word_speakers(
                diarize_segments, 
                transcription_result
                )                  
              
            return result
        
        # Start and time transcription
        logging.info(f"👨‍💻 Transcription started with whisper🆇 for {filepath.name}")
        t_start = time.time()
        
        # Run the transcription
        transcription_task = partial(
            whisperx_task, 
            task='transcribe', 
            language=self.file_language
            )
        transcription_result = little_helper.run_with_progress(
            description=f"[cyan]→ Transcribing ({'CUDA' if self.device == 'cuda:0' else 'CPU'}) [bold]{filepath.name}",
            task=transcription_task
        )
        
        # Speaker annotation
        if self.annotate:
            annotation_task = partial(
                whisperx_annotation, 
                transcription_result
                )
            transcription_result = little_helper.run_with_progress(
            description=f"[purple]→ Annotating ({self.device.upper()}) [bold]{filepath.name}",
            task=annotation_task
        )
        
        # Fill in missing timestamps and adjust word chunk length
        transcription_result['segments'] = fill_missing_timestamps(
            transcription_result['segments']
            )
        transcription_result = self.adjust_word_chunk_length(
            transcription_result
            )
        
        # Create result dict and append transcription to it
        result = {'transcriptions': {}}
        result['transcriptions'][self.file_language] = transcription_result
        
        if self.verbose:
            print(f"{result['transcriptions'][self.file_language]['text']}")
        
        # Translation task (to English)
        if self.translate and self.file_language != 'en':
            translation_task = partial(
                whisperx_task, 
                task='translate', 
                language='en'
                )
            translation_result = little_helper.run_with_progress(
                description=f"[dark_blue]→ Translating ({'CUDA' if self.device == 'cuda:0' else 'CPU'}) [bold]{filepath.name}",
                task=translation_task
            )
            
            # Speaker annotation
            if self.annotate:
                annotation_task = partial(
                    whisperx_annotation, 
                    translation_result
                    )
                translation_result = little_helper.run_with_progress(
                description=f"[purple]→ Annotating ({self.device.upper()}) [bold]{filepath.name}",
                task=annotation_task
            )
            
            # Fill in missing timestamps and adjust word chunk length
            translation_result['segments'] = fill_missing_timestamps(
                translation_result['segments']
                )
            translation_result = self.adjust_word_chunk_length(
                translation_result
                )
            result['transcriptions']['en'] = translation_result
            
            if self.verbose:
                print(f"{result['transcriptions']['en']['text']}")

        # Create full transcription with speaker annotation
        if self.annotate:
            result = self.create_text_with_speakers(result)
        
        logging.info(f"👨‍💻 Transcription completed in {time.time() - t_start:.2f} sec.")
        
        return {'transcription': result}

    def transcribe_with_insane_whisper(self, filepath: Path) -> dict:
        """
        Transcribes a file using the 'insanely-fast-whisper' implementation:
        https://github.com/Vaibhavs10/insanely-fast-whisper

        This method utilizes the 'insanely-fast-whisper' implementation of 
        OpenAI Whisper for automatic speech recognition. It initializes a 
        pipeline for transcription and retrieves the result. If speaker 
        detection is enabled, it also annotates speakers in the transcription 
        result.

        Parameters:
        - filepath (Path): The path to the audio file for transcription.

        Returns:
        - dict: A dictionary containing the transcription result and, if 
                speaker detection is enabled, the speaker diarization result. 
                The transcription result includes the recognized text and 
                timestamps if available.
        """
        import torch
        from transformers import pipeline
        from transformers.utils import is_flash_attn_2_available
        from whisply import diarize_utils
        
        def insane_whisper_annotation(transcription_result: dict) -> dict:
            # Speaker annotation
            annotation_result = diarize_utils.diarize(
                transcription_result,
                diarization_model='pyannote/speaker-diarization-3.1',
                hf_token=self.hf_token,
                file_name=str(filepath),
                num_speakers=self.num_speakers,
                min_speakers=None,
                max_speakers=None,
            )
            # Transform annotation_result to correct dict structure                
            transcription_result = self.to_transcription_dict(annotation_result)
            return transcription_result

        # Start and time transcription
        logging.info(
            f"👨‍💻 Transcription started with 🚅 insane-whisper for {filepath.name}"
            )
        t_start = time.time()
        
        try:
            pipe = pipeline(
                "automatic-speech-recognition",
                model = self.model, 
                torch_dtype = torch.float32 if self.device == 'cpu' else torch.float16,
                device = self.device,
                model_kwargs = {
                    "attn_implementation": "flash_attention_2"
                    } if is_flash_attn_2_available() else {
                        "attn_implementation": "eager"
                        },
            )
            
            # Define transcription function
            def transcription_task():
                transcription_result = pipe(
                    str(filepath),
                    chunk_length_s = 30,
                    batch_size = 8,
                    return_timestamps = 'word', # True, word, chunk
                    generate_kwargs = {
                        'language': self.file_language,
                        }
                )
                return transcription_result
             
            # Transcription
            transcription_result = little_helper.run_with_progress(
                description=f"[cyan]→ Transcribing ({self.device.upper()}) [bold]{filepath.name}",
                task=transcription_task
            )
            
            # Speaker annotation
            if self.annotate:
                transcription_result = little_helper.run_with_progress(
                    description=f"[purple]→ Annotating ({self.device.upper()}) [bold]{filepath.name}",
                    task=partial(
                        insane_whisper_annotation,
                        transcription_result
                    )
                )
                            
            # Adjust word chunk length
            transcription_result = self.to_whisperx(transcription_result)
            transcription_result = self.adjust_word_chunk_length(transcription_result)
            
            # Build result dict            
            result = {'transcriptions': {}}
            result['transcriptions'] = {
                self.file_language: transcription_result
            }
    
            if self.verbose:
                print(result['transcriptions'][self.file_language]['text'])
            
            # Translation
            if self.translate and self.file_language != 'en':
                # Define translation task
                def translation_task():
                    translation_result = pipe(
                        str(filepath),
                        chunk_length_s = 30,
                        batch_size = 8 if self.device in ['cpu', 'mps'] else 24,
                        return_timestamps = 'word',
                        generate_kwargs = {'task': 'translate',
                                           'language': 'en'}
                    )
                    return translation_result
                
                # Run the translation task
                translation_result = little_helper.run_with_progress(
                    description=f"[dark_blue]→ Translating ({self.device.upper()}) [bold]{filepath.name}",
                    task=translation_task
                )
                
                # Speaker annotation
                if self.annotate:
                    translation_result = little_helper.run_with_progress(
                        description=f"[purple]→ Annotating ({self.device.upper()}) [bold]{filepath.name}",
                        task=partial(
                            insane_whisper_annotation,
                            translation_result
                        )
                    )
                
                # Adjust word chunk length
                translation_result = self.to_whisperx(translation_result)
                translation_result = self.adjust_word_chunk_length(translation_result)
                
                result['transcriptions']['en'] = translation_result
                
                if self.verbose:
                    print(result['transcriptions']['en']['text'])
                    
            if self.annotate:
                # Create full transcription with speaker annotation
                result = self.create_text_with_speakers(result)

        except Exception as e:
            print(f'{e}')
        
        # Stop timing transcription
        logging.info(f"👨‍💻 Transcription completed in {time.time() - t_start:.2f} sec.")
        
        return {'transcription': result}

    def transcribe_with_faster_whisper(
        self, 
        filepath: Path, 
        num_workers: int = 1
        ) -> dict:
        """
        Transcribes an audio file using the 'faster-whisper' implementation: 
        https://github.com/SYSTRAN/faster-whisper

        This method utilizes the 'faster-whisper' implementation of OpenAI 
        Whisper for automatic speech recognition. It loads the model and sets 
        parameters for transcription. After transcription, it formats the 
        result into segments with timestamps and combines them into a single 
        text. If speaker detection is enabled, it also annotates speakers in 
        the transcription result.

        Parameters:
        - filepath (Path): The path to the audio file for transcription.
        - num_workers (int): The number of workers to use for transcription.

        Returns:
        - dict: A dictionary containing the transcription result and, if 
                speaker detection is enabled, the speaker diarization result. 
                The transcription result includes the recognized text and 
                segmented chunks with timestamps if available.
        """
        from faster_whisper import WhisperModel, BatchedInferencePipeline

        # Start and time transcription
        logging.info(f"👨‍💻 Transcription started with 🏃‍♀️‍➡️ faster-whisper for {filepath.name}")
        t_start = time.time()
        
        # Load model and set parameters
        model = BatchedInferencePipeline(
            model = WhisperModel(
            self.model, 
            device='cpu' if self.device in ['mps', 'cpu'] else 'cuda', 
            num_workers=num_workers, 
            compute_type='int8' if self.device in ['mps', 'cpu'] else 'float16'
        ))
        
        # Define the transcription task
        def transcription_task():
            segments, _ = model.transcribe(
                str(filepath), 
                beam_size=5, 
                language=self.file_language,
                word_timestamps=True,
                batch_size=16
            )
                
            chunks = []    
            for segment in segments:
                seg = {
                    'timestamp': (float(f"{segment.start:.2f}"), float(f"{segment.end:.2f}")),
                    'text': segment.text.strip(),
                    'words': [{
                        'word': i.word.strip(),
                        'start': float(f"{i.start:.2f}"),
                        'end': float(f"{i.end:.2f}"),
                        'score': float(f"{i.probability:.2f}")
                    } for i in segment.words]
                }
                chunks.append(seg)
                
                # If verbose Flag 
                if self.verbose:
                    print(seg['text'])
                    
            return chunks
        
        # Add progress bar and run the transcription task
        chunks = little_helper.run_with_progress(
            description=f"[cyan]→ Transcribing ({self.device.upper()}) [bold]{filepath.name}",
            task=transcription_task
        )
        
        # Create result dict and append transcriptions to it
        result = {'transcriptions': {}}
        result['transcriptions'][self.file_language] = {
                'text': ' '.join([segment['text'].strip() for segment in chunks]),
                'chunks': chunks
                }

        # Translation
        if self.translate and self.file_language != 'en':
            # Define the translation task
            def translation_task():
                segments, _ = model.transcribe(
                    str(filepath), 
                    beam_size=5, 
                    task='translate', 
                    language='en',
                    word_timestamps=True)
                
                translation_chunks = []    
                for segment in segments:
                    seg = {
                        'timestamp': (float(f"{segment.start:.2f}"), float(f"{segment.end:.2f}")),
                        'text': segment.text.strip(),
                        'words': [{
                            'word': i.word.strip(),
                            'start': float(f"{i.start:.2f}"),
                            'end': float(f"{i.end:.2f}"),
                            'score': float(f"{i.probability:.2f}")
                        } for i in segment.words]
                    }
                    translation_chunks.append(seg)
                    
                    # If verbose Flag 
                    if self.verbose:
                        print(seg['text'])
                        
                return translation_chunks
            
            # Add progress bar and run the translation task
            translation_chunks = little_helper.run_with_progress(
                description=f"[dark_blue]→ Translating ({self.device.upper()}) [bold]{filepath.name}",
                task=translation_task
            )

            # Add translation to result dict
            result['transcriptions']['en'] = {
                'text': ' '.join([segment['text'].strip() for segment in translation_chunks]),
                'chunks': translation_chunks
                }
        
        # Stop timing transcription
        logging.info(f"👨‍💻 Transcription completed in {time.time() - t_start:.2f} sec.")
        
        return {'transcription': result}
   
    def detect_language(self, filepath, audio_array) -> str:   
        """
        Detects the language of the input file.
        """
        from faster_whisper import WhisperModel
        
        logging.info(f"Detecting language of file: {filepath.name}")    
        
        def run_language_detection():
            lang_detection_model = WhisperModel(
                models.set_supported_model(self.model_provided, implementation='faster-whisper'), 
                device='cpu' if self.device in ['mps', 'cpu'] else 'cuda', 
                compute_type='int8' if self.device in ['mps', 'cpu'] else 'float16'
                )
            lang, score, _ = lang_detection_model.detect_language(audio_array)
            return lang, score
        
        lang, score = little_helper.run_with_progress(
            description=f"[dark_goldenrod]→ Detecting language for [bold]{filepath.name}",
            task=run_language_detection                  
        )    
        
        self.file_language = lang   

        print(f'→ Detected language "{lang}" with probability {score:.2f}')
        logging.info(f'Detected language → "{lang}" with probability {score:.2f}')
        
    def process_files(self, files) -> None:
        """
        Processes a list of audio files for transcription and/or diarization.

        This method logs the processing parameters, extracts filepaths from the
        input list, and initializes an empty list for storing results. Each 
        file is processed based on the compute device specified ('mps', 'cuda:0',
        or 'cpu'). Appropriate transcription method is chosen based on the 
        device. Results, including file ids, paths, transcriptions, and 
        diarizations, are stored in a dictionary and saved to a designated 
        output directory. Each result is also appended to `self.processed_files`.

        Parameters:
        files (list of str): A list of file paths or file-like objects 
        representing the audio files to be processed.
        """
        logging.info(f"Provided parameters for processing: {self.metadata}")
        
        # Get filepaths
        filepath_handler = FilePathProcessor(self.file_formats)
        [filepath_handler.get_filepaths(f) for f in files]            
        self.filepaths = filepath_handler.filepaths
        
        # Process filepaths
        logging.info(f"Processing files: {self.filepaths}")
        
        self.processed_files = []
        for idx, filepath in enumerate(self.filepaths):      
                  
            # Create and set output_dir and output_filepath
            self.output_dir = little_helper.set_output_dir(filepath, self.base_dir)
            output_filepath = self.output_dir / Path(filepath).stem
            
            # Convert file format 
            filepath, audio_array = little_helper.check_file_format(
                filepath=filepath,
                del_originals=self.del_originals
                )
            
            # Detect file language
            if not self.file_language:
                self.detect_language(filepath, audio_array)

            logging.info(f"Transcribing file: {filepath.name}")
            
            # Transcription and speaker annotation
            if self.device == 'mps':
                self.model = models.set_supported_model(
                    self.model_provided, 
                    implementation='insane-whisper'
                )
                print(f'[deep_pink4]→ Using {self.device.upper()} and 🚅 Insanely-Fast-Whisper with model "{self.model}"')
                result_data = self.transcribe_with_insane_whisper(filepath)
            
            elif self.device in ['cpu', 'cuda:0']:
                if self.annotate or self.subtitle:
                    # WhisperX for annotation / subtitling
                    self.model = models.set_supported_model(
                        self.model_provided, 
                        implementation='whisperx'
                    )
                    print(f'[deep_pink4]→ Using {self.device.upper()} and whisper🆇  with model "{self.model}"')
                    result_data = self.transcribe_with_whisperx(filepath)
                else:
                    # Faster-Whisper for raw transcription
                    self.model = models.set_supported_model(
                        self.model_provided, 
                        implementation='faster-whisper'
                    )
                    print(f'[deep_pink4]→ Using {self.device.upper()} and 🏃‍♀️‍➡️ Faster-Whisper with model "{self.model}"')
                    result_data = self.transcribe_with_faster_whisper(filepath)

            result = {
                'id': f'file_00{idx + 1}',
                'created': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'input_filepath': str(filepath.absolute()),
                'output_filepath': str(Path(output_filepath).absolute()),
                'written_files': None,
                'device': self.device,
                'model': self.model,
                'transcription': result_data['transcription']['transcriptions'],
            }

            # Save results
            result['written_files'] = output_utils.OutputWriter(
                corrections=self.corrections
                ).save_results(
                    result=result,
                    export_formats=self.export_formats
                    )
            
            self.processed_files.append(result)
            
            if not self.file_language_provided:
                self.file_language = None



import re
import os
import json
import logging
import ffmpeg
import validators
import numpy as np

from enum import Enum
from pathlib import Path
from typing import Callable, Any, List
from rich import print
from rich.progress import Progress, TimeElapsedColumn, TextColumn, SpinnerColumn
from whisply import download_utils

# Set logging configuration
logger = logging.getLogger('little_helper')
logger.setLevel(logging.INFO)


class DeviceChoice(str, Enum):
    AUTO = 'auto'
    CPU = 'cpu'
    GPU = 'gpu'
    MPS = 'mps'
    
    
def get_device(device: DeviceChoice = DeviceChoice.AUTO) -> str:
    """
    Determine the computation device based on user preference and 
    availability.
    """
    import torch
    
    if device == DeviceChoice.AUTO:
        if torch.cuda.is_available():
            device = 'cuda:0'
        elif torch.backends.mps.is_available():
            device = 'mps'
        else:
            device = 'cpu'
    elif device == DeviceChoice.GPU:
        if torch.cuda.is_available():
            device = 'cuda:0'
        else:
            device = 'cpu'
    elif device == DeviceChoice.MPS:
        if torch.backends.mps.is_available():
            device = 'mps'
        else:
            device = 'cpu'
    elif device == DeviceChoice.CPU:
        device = 'cpu'
    else:
        device = 'cpu'
    return device


class FilePathProcessor:
    """
    Utility class for validating various filepaths.
    """
    def __init__(self, file_formats: List[str]):
        self.file_formats = [fmt.lower() for fmt in file_formats]
        self.filepaths: List[Path] = []

    def get_filepaths(self, filepath: str):
        """
        Processes the provided filepath which can be a URL, a single file, a directory,
        or a .list file containing multiple paths/URLs. It validates each input, downloads
        URLs if necessary, and accumulates valid file paths for further processing.
        """
        path = Path(filepath).expanduser().resolve()
        
        try:
            # Handle URL
            if validators.url(filepath):
                logging.info(f"Processing URL: {filepath}")
                downloaded_path = download_utils.download_url(
                    filepath, 
                    downloads_dir=Path('./downloads')
                )
                if downloaded_path:
                    self.filepaths.append(downloaded_path)
                else:
                    logging.error(f"Failed to download URL: {filepath}")
                    print(f"→ Failed to download URL: {filepath}")
                return  

            # Handle .list file
            elif path.suffix.lower() == '.list':
                if not path.is_file():
                    logging.error(f'The .list file "{path}" does not exist or is not a file.')
                    print(f'→ The .list file "{path}" does not exist or is not a file.')
                    return
                
                logging.info(f"Processing .list file: {path}")
                with path.open('r', encoding='utf-8') as file:
                    lpaths = set()
                    for line in file:
                        lpath = line.strip()
                        if not lpath:
                            continue 
                        lpaths.add(lpath)
                        
                    for lpath in lpaths:
                        if validators.url(lpath):
                            downloaded_path = download_utils.download_url(
                                lpath, 
                                downloads_dir=Path('./downloads')
                            )
                            if downloaded_path:
                                self.filepaths.append(downloaded_path)
                            else:
                                print(f'→ Failed to download URL: {lpath}')
                        else:
                            self._process_path(lpath)
                return

            # Handle single file or directory
            else:
                self._process_path(path)

        except Exception as e:
            logging.exception(f"An unexpected error occurred while processing '{filepath}': {e}")
            return

        # Remove duplicates by converting to a set of resolved absolute paths
        unique_filepaths = set(p.resolve() for p in self.filepaths)
        self.filepaths = list(unique_filepaths)

        # Filter out files that have already been converted
        self._filter_converted_files()
        
        # Final check to ensure there are files to process
        if not self.filepaths:
            logging.warning(f'No valid files found for processing. Please check the provided path: "{filepath}".')
            print(f'→ No valid files found for processing. Please check the provided path: "{filepath}".')
        else:
            logging.info(f"Total valid files to process: {len(self.filepaths)}")

    def _process_path(self, path_input: str | Path):
        """
        Processes a single path input, which can be a file or a directory.
        """
        path = Path(path_input).expanduser().resolve()

        if path.is_file():
            if path.suffix.lower() in self.file_formats:
                logging.info(f"Adding file: {path}")
                normalized_path = self._normalize_filepath(path)
                self.filepaths.append(normalized_path)
            else:
                logging.warning(f'File "{path}" has unsupported format and will be skipped.')
                print(f'→ File "{path}" has unsupported format and will be skipped.')
        elif path.is_dir():
            logging.info(f"Processing directory: {path}")
            for file_format in self.file_formats:
                for file in path.rglob(f'*{file_format}'):
                    if file.is_file():
                        logging.debug(f"Found file: {file}")
                        normalized_path = self._normalize_filepath(file)
                        self.filepaths.append(normalized_path)
        else:
            logging.error(f'Path "{path}" does not exist or is not accessible.')
            print(f'→ Path "{path}" does not exist or is not accessible.')
             
    def _normalize_filepath(self, filepath: Path) -> Path:
        """
        Normalizes the filepath by replacing non-word characters with underscores,
        collapsing multiple underscores into one, and removing leading/trailing underscores.
        """
        new_filename = re.sub(r'\W+', '_', filepath.stem)
        new_filename = new_filename.strip('_')
        new_filename = re.sub(r'_+', '_', new_filename)
        
        suffix = filepath.suffix.lower()
        
        # Construct the new path
        new_path = filepath.parent / f"{new_filename}{suffix}"
        
        # Rename the file
        filepath.rename(new_path)

        return new_path.resolve()

    def _filter_converted_files(self):
        """
        Removes files that have already been converted to avoid redundant processing.
        """
        converted_suffix = '_converted.wav'
        original_filepaths = []
        converted_filepaths = set()

        for fp in self.filepaths:            
            if fp.name.endswith(converted_suffix):
                converted_filepaths.add(fp)
            else:
                original_filepaths.append(fp)
                
        # Remove originals if their converted version exists
        filtered_filepaths = [
            fp for fp in original_filepaths
            if not (fp.with_name(fp.stem + converted_suffix) in converted_filepaths)
        ]
        
        # Extened filtered paths with converted paths
        filtered_filepaths.extend(converted_filepaths)

        removed_count = len(self.filepaths) - len(filtered_filepaths)
        if removed_count > 0:
            logging.info(f"Removed {removed_count} files already converted.")
        self.filepaths = filtered_filepaths

def ensure_dir(dir: Path) -> None:
    if not dir.exists():
        dir.mkdir(parents=True)
    return dir
        
def set_output_dir(filepath: Path, base_dir: Path) -> None:
    output_dir = base_dir / filepath.stem
    ensure_dir(output_dir)
    return output_dir

def return_valid_fileformats() -> list[str]:
    return [
        '.mp3',
        '.wav',
        '.m4a',
        '.aac',
        '.flac',
        '.ogg',
        '.mkv',
        '.mov',
        '.mp4',
        '.avi',
        '.mpeg',
        '.vob'
        ]

def check_file_format(
    filepath: Path, 
    del_originals: bool = True
    ) -> tuple[Path, np.ndarray]:
    """
    Checks the format of an audio file and converts it if it doesn't meet specified criteria.
    Then, loads the audio into a 1D NumPy array.

    The function uses `ffmpeg` to probe the metadata of an audio file at the given `filepath`.
    It checks if the audio stream meets the following criteria:
    - Codec name: 'pcm_s16le'
    - Sample rate: 16000 Hz
    - Number of channels: 1 (mono)

    If the audio stream does not meet these criteria, the function attempts to convert the file
    to meet the required format and saves the converted file with a '_converted.wav' suffix in the same directory.
    After successful conversion, it deletes the original file.

    Finally, it loads the audio (original or converted) as a 1D NumPy array and returns it.

    Args:
        filepath (Path): The path to the audio file to be checked and potentially converted.

    Returns:
        filepath (Path): filepath of the checked and / or converted audio file.
        np.ndarray: 1D NumPy array of the audio data.
    """ 
    import librosa
    
    # Define the converted file path
    new_filepath = filepath.with_name(f"{filepath.stem}_converted.wav")
    
    converted = False
    
    if new_filepath.exists():
        target_filepath = new_filepath
        converted = True
    else:
        try:
            # Probe the audio file for stream information
            probe = ffmpeg.probe(str(filepath))
            audio_streams = [stream for stream in probe['streams'] if stream['codec_type'] == 'audio']
            
            if not audio_streams:
                raise ValueError(f"→ No audio stream found for {filepath}. Please check if the file you have provided contains audio content.")
            
            audio_stream = audio_streams[0]
            codec_name = audio_stream.get('codec_name')
            sample_rate = int(audio_stream.get('sample_rate', 0))
            channels = int(audio_stream.get('channels', 0))
            
            # Check if the audio stream meets the criteria
            if codec_name != 'pcm_s16le' or sample_rate != 16000 or channels != 1:
                try:
                    # Convert the file and show progress
                    run_with_progress(
                        description=(
                            f"[orchid]→ Converting file to .wav: {filepath.name}"
                            ), 
                        task=lambda: convert_file_format(
                            old_filepath=filepath, 
                            new_filepath=new_filepath
                            )
                    )
                    target_filepath = new_filepath
                    converted = True
                except Exception as e:
                    raise RuntimeError(
                        f"→ An error occurred while converting {filepath}: {e}"
                        )
            else:
                # If already in correct format, use the original file
                target_filepath = filepath
            
        except ffmpeg.Error as e:
            print(f"→ Error running ffprobe: {e}")
            print(f"→ You may have provided an unsupported file type.\
Please check 'whisply --list_formats' for all supported formats.")
    
    try:
        # Load the audio file into a NumPy array
        audio, _ = librosa.load(str(target_filepath), sr=16000, mono=True)
        audio_array = audio.astype(np.float32)
    except Exception as e:
        raise RuntimeError(f"Failed to load audio from {target_filepath}: {e}") from e
    
    # If conversion occurred delete the original file if del_originals
    if (converted and del_originals) and target_filepath != filepath:
        try:
            os.remove(filepath)
        except OSError as e:
            print(f"Warning: {e}")
    
    return Path(target_filepath), audio_array
       
def convert_file_format(old_filepath: str, new_filepath: str):
    """
    Converts a video file into an audio file in WAV format using the ffmpeg library.
    """
    (
        ffmpeg
        .input(str(old_filepath))
        .output(str(new_filepath), 
                acodec='pcm_s16le', # Audio codec: PCM signed 16-bit little-endian
                ar='16000',         # Sampling rate 16 KHz
                ac=1)               # Mono channel                       
        .run(quiet=True,
             overwrite_output=True)
    )

def load_config(config: json) -> dict:
    with open(config, 'r', encoding='utf-8') as file:
        return json.load(file)

def format_time(seconds, delimiter=',') -> str:
    """
    Function for time conversion.
    """
    h = int(seconds // 3600)
    m = int((seconds % 3600) // 60)
    s = int(seconds % 60)
    ms = int((seconds - int(seconds)) * 1000)
    
    return f"{h:02}:{m:02}:{s:02}{delimiter}{ms:03}"
 
def run_with_progress(description: str, task: Callable[[], Any]) -> Any:
    """
    Helper function to run a task with a progress bar.
    """
    with Progress(
        SpinnerColumn(),
        TimeElapsedColumn(),
        TextColumn("[progress.description]{task.description}")
    ) as progress:
        progress.add_task(description, total=None)
        return task()


import os
import typer
import warnings

from pathlib import Path
from typing import Optional, List
from rich import print
from whisply import output_utils, little_helper
from whisply import post_correction as post
from whisply.output_utils import ExportFormats
from whisply.little_helper import DeviceChoice

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=UserWarning)

app = typer.Typer()

@app.command(no_args_is_help=True)
def main(
    files: Optional[List[str]] = typer.Option(
        None,
        "--files",
        "-f",
        help="Path to file, folder, URL or .list to process.",
    ),
    output_dir: Path = typer.Option(
        Path("./transcriptions"),
        "--output_dir",
        "-o",
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Folder where transcripts should be saved.",
    ),
    device: DeviceChoice = typer.Option(
        DeviceChoice.AUTO,
        "--device",
        "-d",
        help="Select the computation device: CPU, GPU (NVIDIA), or MPS (Mac M1-M4).",
    ),
    model: str = typer.Option(
        "large-v3-turbo",
        "--model",
        "-m",
        help='Whisper model to use (List models via --list_models).',
    ),
    lang: Optional[str] = typer.Option(
        None,
        "--lang",
        "-l",
        help='Language of provided file(s) ("en", "de") (Default: auto-detection).',
    ),
    annotate: bool = typer.Option(
        False,
        "--annotate",
        "-a",
        help="Enable speaker annotation (Saves .rttm | Default: False).",
    ),
    num_speakers: Optional[int] = typer.Option(
        None,
        "--num_speakers",
        "-num",
        help="Number of speakers to annotate (Default: auto-detection).",
    ),
    hf_token: Optional[str] = typer.Option(
        None,
        "--hf_token",
        "-hf",
        help="HuggingFace Access token required for speaker annotation.",
    ),
    subtitle: bool = typer.Option(
        False,
        "--subtitle",
        "-s",
        help="Create subtitles (Saves .srt, .vtt and .webvtt | Default: False).",
    ),
    sub_length: int = typer.Option(
        5,
        "--sub_length",
        help="Subtitle segment length in words."
    ),
    translate: bool = typer.Option(
        False,
        "--translate",
        "-t",
        help="Translate transcription to English (Default: False).",
    ),
    export_format: ExportFormats = typer.Option(
        ExportFormats.ALL,
        "--export",
        "-e",
        help="Choose the export format."
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Print text chunks during transcription (Default: False).",
    ),
    del_originals: bool = typer.Option(
        False,
        "--del_originals",
        "-del",
        help="Delete original input files after file conversion. (Default: False)",
    ),
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        help="Path to configuration file.",
    ),
    post_correction: Optional[Path] = typer.Option(
        None,
        "--post_correction",
        "-post",
        help="Path to YAML file for post-correction.",
    ),
    list_filetypes: bool = typer.Option(
        False,
        "--list_filetypes",
        help="List supported audio and video file types.",
    ),
    list_models: bool = typer.Option(
        False,
        "--list_models",
        help="List available models.",
    ),
):
    """
    WHISPLY 💬 Transcribe, translate, annotate and subtitle audio and video files with OpenAI's Whisper ... fast!
    """
    from whisply import little_helper, transcription, models

    # Load configuration from config.json if provided
    if config:
        config_data = little_helper.load_config(config)
        files = files or Path(config_data.get("files")) if config_data.get("files") else files
        output_dir = Path(config_data.get("output_dir")) if config_data.get("output_dir") else output_dir
        device = DeviceChoice(config_data.get("device", device.value))
        model = config_data.get("model", model)
        lang = config_data.get("lang", lang)
        annotate = config_data.get("annotate", annotate)
        num_speakers = config_data.get("num_speakers", num_speakers)
        translate = config_data.get("translate", translate)
        hf_token = config_data.get("hf_token", hf_token)
        subtitle = config_data.get("subtitle", subtitle)
        sub_length = config_data.get("sub_length", sub_length)
        verbose = config_data.get("verbose", verbose)
        del_originals = config_data.get("del_originals", del_originals)
        post_correction = config_data.get("post_correction", post_correction)

    # Print supported filetypes 
    if list_filetypes:
        supported_filetypes = "Supported filetypes: "
        supported_filetypes += ' '.join(little_helper.return_valid_fileformats())
        print(f"{supported_filetypes}")
        raise typer.Exit()

    # Print available models
    if list_models:
        available_models = "Available models:\n... "
        available_models += '\n... '.join(models.WHISPER_MODELS.keys())
        print(f"{available_models}")
        raise typer.Exit()

    # Check if provided model is available
    if not models.ensure_model(model):
        msg = f"""→ Model "{model}" is not available.\n→ Available models:\n... """
        msg += '\n... '.join(models.WHISPER_MODELS.keys())
        print(f"{msg}")
        raise typer.Exit()

    # Check for HuggingFace Access Token if speaker annotation is enabled
    if annotate and not hf_token:
        hf_token = os.getenv('HF_TOKEN')
        if not hf_token:
            print('→ Please provide a HuggingFace access token (--hf_token / -hf) to enable speaker annotation.')
            raise typer.Exit()

    # Determine the computation device
    device_str = little_helper.get_device(device=device)
    
    # Determine the ExportFormats
    export_formats = output_utils.determine_export_formats(
        export_format, 
        annotate, subtitle
        )
    
    # Load corrections if post_correction is provided
    if post_correction:
        corrections = post.load_correction_list(post_correction)
    
    # Instantiate TranscriptionHandler
    service = transcription.TranscriptionHandler(
        base_dir=output_dir,
        device=device_str,
        model=model,
        file_language=lang, 
        annotate=annotate, 
        num_speakers=num_speakers,
        translate=translate,
        hf_token=hf_token, 
        subtitle=subtitle,
        sub_length=sub_length,
        verbose=verbose,
        del_originals=del_originals,
        corrections=corrections if post_correction else None,
        export_formats=export_formats
    )
    # Process files
    service.process_files(files)

def run():
    app()

if __name__ == "__main__":
    run()



def transform2gat2(result: dict) -> dict:
    """
    Transforms the 'result' dictionary by inserting pause dictionaries into the words list
    where the pause duration between words falls within specified ranges.

    Args:
        result (dict): The original transcription result containing words with timing information.

    Returns:
        dict: The transformed result with pauses inserted.
    """
    # Iterate over each language in transcription
    for _, transcription in result.get('transcriptions').items():
        chunks = transcription.get('chunks')
        
        # Iterate over each chunk
        for chunk in chunks:
            words = chunk.get('words')
            # text = chunk.get('text')
            
            if not words:
                continue  # Skip if there are no words in the chunk
            
            new_words = []
            for i in range(len(words) - 1):
                new_text = ''
                
                current_word = words[i]
                next_word = words[i + 1]
                new_words.append(current_word)
                
                # Calculate the pause duration
                pause_duration = next_word['start'] - current_word['end']
                
                # Skip if pause_duration is non-positive
                if pause_duration <= 0:
                    continue
                
                # Determine the type of pause based on duration
                if 0.2 <= pause_duration < 0.5:
                    pause_marker = "(-)"
                elif 0.5 <= pause_duration < 0.8:
                    pause_marker = "(--)"
                elif 0.8 <= pause_duration < 1:
                    pause_marker = "(---)"
                elif pause_duration >= 1:
                    pause_marker = f"({pause_duration:.2f})"
                else:
                    continue
                
                # Create the pause dictionary
                pause_dict = {
                    "word": pause_marker,
                    "start": current_word.get("end"),
                    "end": next_word.get("start"),
                    "score": 1,
                    "speaker": current_word.get("speaker"),
                }
                new_words.append(pause_dict)
                new_text += f"{pause_dict['word']} "
            
            # Append the last word of the chunk
            if words:
                new_words.append(words[-1])
            
            print(new_words)
            
            # Update the words list with the new_words
            chunk['words'] = new_words
            chunk['text'] = new_text.strip()
    
    return result



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="whisply",
    author="Thomas Schmidt, Renat Shigapov",
    version='0.7.4',
    packages=find_packages(),
    license="Apache-2.0",
    description="Transcribe, translate, annotate and subtitle audio and video files with OpenAI's Whisper ... fast!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tsmdt/whisply",
    install_requires=[
        'typer==0.12.5',
        'numpy==1.26.4',
        'faster-whisper==1.1.0',
        'ffmpeg-python==0.2.0',
        'optimum==1.19.1',
        'pyannote.audio==3.1.1',
        'pyannote.core==5.0.0',
        'pyannote.database==5.1.0',
        'pyannote.metrics==3.2.1',
        'pyannote.pipeline==3.0.1',
        'rich==13.7.1',
        'torch==2.3.0',
        'torch-audiomentations==0.11.1',
        'torch-pitch-shift==1.2.4',
        'torchaudio==2.3.0',
        'torchmetrics==1.3.2',
        'transformers==4.39.3',
        'validators==0.28.1',
        'yt-dlp>=2024.07.07',
        'whisperx',
        'gradio==5.6.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    entry_points={
        'console_scripts': [
            'whisply=whisply.cli:run',
        ],
    },
)



import re
import logging
import yt_dlp as url_downloader

from pathlib import Path
from datetime import datetime

from whisply import little_helper


# Set logging configuration
logger = logging.getLogger('download_utils')
logger.setLevel(logging.DEBUG)


def download_url(url: str, downloads_dir: Path) -> Path:
    """
    Downloads a media file from a specified URL, typically a YouTube URL, extracting audio
    in WAV format, and then renames the file based on the media title. 

    The function first ensures the downloads directory exists, then initiates a download
    using youtube-dl with specific options set for audio quality and format. After the 
    download, it extracts the video's information without downloading it again to 
    rename the file more meaningfully based on the video title. Special characters 
    in the title are replaced with underscores, and unnecessary leading or trailing 
    underscores are removed.

    Args:
        url (str): The URL of the video to download.
        downloads_dir (Path): The directory path where the downloaded file should be stored.

    Returns:
        Path: A path object pointing to the renamed downloaded file. If there is an error 
        during the download or file processing, returns None.

    Raises:
        Exception: Outputs an error message to the console if the download fails.

    Examples:
        >>> download_url("https://www.youtube.com/watch?v=example", Path("/downloads"))
        Path('/downloads/example.wav')
    """
    little_helper.ensure_dir(downloads_dir)
    
    temp_filename = f"temp_{datetime.now().strftime('%Y%m%d_%H_%M_%S')}"
    options = {
        'format': 'bestaudio/best',
        'postprocessors': [{'key': 'FFmpegExtractAudio', 
                            'preferredcodec': 'wav', 
                            'preferredquality': '192'}],
        'outtmpl': f'{downloads_dir}/{temp_filename}.%(ext)s'
    }
    try:
        with url_downloader.YoutubeDL(options) as ydl:
            logger.debug(f"Downloading {url}")
            
            # Download url
            ydl.download([url])
            video_info = ydl.extract_info(url, download=False)
            downloaded_file = list(downloads_dir.glob(f'{temp_filename}*'))[0]
            logger.debug(f"Download complete for {downloaded_file}")
            
            # Normalize title
            new_filename = re.sub(r'\W+', '_', video_info.get('title', 'downloaded_video'))
            
            # Remove trailing underscores
            if new_filename.startswith('_'):
                new_filename = new_filename[1:]
            if new_filename.endswith('_'):
                new_filename = new_filename[:-1]
            
            # Rename the file
            renamed_file = downloaded_file.rename(f"{downloads_dir}/{new_filename}{downloaded_file.suffix}")
            logger.debug(f"Renamed downloaded file to {renamed_file}")
            return Path(renamed_file)
        
    except Exception as e:
        print(f'Error downloading {url}: {e}')
        return None
    


import re
import json
import logging
import typer

from enum import Enum
from pathlib import Path
from typing import List, Dict, Tuple
from rich import print
from whisply import little_helper
from whisply.post_correction import Corrections

# Set logging configuration
logger = logging.getLogger('little_helper')
logger.setLevel(logging.INFO)


class ExportFormats(str, Enum):
    ALL = 'all'
    JSON = 'json'
    TXT = 'txt'
    RTTM = 'rttm'
    VTT = 'vtt'
    WEBVTT = 'webvtt'
    SRT = 'srt'
    
    
def determine_export_formats(
    export_format: ExportFormats,
    annotate: bool,
    subtitle: bool
) -> List[str]:
    """
    Determine the export formats based on user options and availability.

    Returns a list of export format strings to be used.
    """
    available_formats = set()
    if export_format == ExportFormats.ALL:
        available_formats.add(ExportFormats.JSON.value)
        available_formats.add(ExportFormats.TXT.value)
        if annotate:
            available_formats.add(ExportFormats.RTTM.value)
        if subtitle:
            available_formats.add(ExportFormats.WEBVTT.value)
            available_formats.add(ExportFormats.VTT.value)
            available_formats.add(ExportFormats.SRT.value)
    else:
        if export_format in (ExportFormats.JSON, ExportFormats.TXT):
            available_formats.add(export_format.value)
        elif export_format == ExportFormats.RTTM:
            if annotate:
                available_formats.add(export_format.value)
            else:
                print("→ RTTM export format requires annotate option to be True.")
                raise typer.Exit()
        elif export_format in (
            ExportFormats.VTT,
            ExportFormats.SRT,
            ExportFormats.WEBVTT
            ):
            if subtitle:
                available_formats.add(export_format.value)
            else:
                print(f"→ {export_format.value.upper()} export format requires subtitle option to be True.")
                raise typer.Exit()
        else:
            print(f"→ Unknown export format: {export_format.value}")
            raise typer.Exit()

    return list(available_formats)


class OutputWriter:
    """
    Class for writing various output formats to disk.
    """
    def __init__(
        self, 
        corrections: Corrections = None
        ):
        self.cwd = Path.cwd()
        self.corrections = corrections
        self.compiled_simple_patterns = (
            self._compile_simple_patterns() if self.corrections else {}
            )
        self.compiled_regex_patterns = (
            self._compile_regex_patterns() if self.corrections else []
            )
    
    def _compile_simple_patterns(self) -> List[Tuple[re.Pattern, str]]:
        """
        Pre-compile regex patterns for simple word corrections.
        Returns a list of tuples containing compiled patterns and their replacements.
        """
        patterns = []
        for wrong, correct in self.corrections.simple.items():
            # Wrap simple corrections with word boundaries
            pattern = re.compile(
                r'\b{}\b'.format(re.escape(wrong)), flags=re.IGNORECASE
                )
            patterns.append((pattern, correct))
            logger.debug(
                f"Compiled simple pattern: '\\b{wrong}\\b' → '{correct}'"
                )
        return patterns
    
    def _compile_regex_patterns(self) -> List[Tuple[re.Pattern, str]]:
        """
        Pre-compile regex patterns for pattern-based corrections.
        Returns a list of tuples containing compiled regex patterns and their replacements.
        """
        patterns = []
        for entry in self.corrections.patterns:
            original_pattern = entry['pattern']
            replacement = entry['replacement']
            
            # Wrap patterns with word boundaries and non-capturing group
            new_pattern = r'\b(?:' + original_pattern + r')\b'
            regex = re.compile(new_pattern, flags=re.IGNORECASE)
            
            patterns.append((regex, replacement))
            logger.debug(
                f"Compiled pattern-based regex: '{new_pattern}' → '{replacement}'"
                )
        return patterns
    
    def correct_transcription(self, transcription: str) -> str:
        """
        Apply both simple and pattern-based corrections to the transcription.
        """
        # Apply simple corrections
        for pattern, correct in self.compiled_simple_patterns:
            transcription = pattern.sub(
                lambda m: self.replace_match(m, correct), transcription
                )
        
        # Apply pattern-based corrections
        for regex, replacement in self.compiled_regex_patterns:
            transcription = regex.sub(replacement, transcription)
        
        return transcription
    
    @staticmethod
    def replace_match(match, correct: str) -> str:
        """
        Replace the matched word while preserving the original casing.
        """
        word = match.group()
        if word.isupper():
            return correct.upper()
        elif word[0].isupper():
            return correct.capitalize()
        else:
            return correct
        
    def _save_file(
        self, 
        content: str, 
        filepath: Path, 
        description: str, 
        log_message: str
        ) -> None:
        """
        Generic method to save content to a file.
        """
        with open(filepath, 'w', encoding='utf-8') as file:
            file.write(content)
        print(f'[blue1]→ Saved {description}: [bold]{filepath.relative_to(self.cwd)}')
        logger.info(f'{log_message} {filepath}')

    def save_json(
        self, 
        result: dict, 
        filepath: Path
        ) -> None:
        with open(filepath, 'w', encoding='utf-8') as fout:
            json.dump(result, fout, indent=4)
        print(f'[blue1]→ Saved .json: [bold]{filepath.relative_to(self.cwd)}')
        logger.info(f"Saved .json to {filepath}")
        
    def save_txt(
        self, 
        transcription: Dict[str, str], 
        filepath: Path
        ) -> None:
        """
        Save the transcription as a TXT file after applying corrections.
        """
        original_text = transcription.get('text', '').strip()
        corrected_text = self.correct_transcription(original_text)
        self._save_file(
            content=corrected_text,
            filepath=filepath,
            description='.txt',
            log_message='Saved .txt transcript to'
        )
    
    def save_txt_with_speaker_annotation(
        self, 
        annotated_text: str, 
        filepath: Path
        ) -> None:
        """
        Save the annotated transcription as a TXT file after applying corrections.
        """
        corrected_annotated_text = self.correct_transcription(annotated_text)
        self._save_file(
            content=corrected_annotated_text,
            filepath=filepath,
            description='.txt with speaker annotation',
            log_message='Saved .txt transcription with speaker annotation →'
        )
    
    def save_subtitles(
        self, 
        text: str, 
        type: str, 
        filepath: Path
        ) -> None:
        """
        Save subtitles in the specified format after applying corrections.
        """
        corrected_text = self.correct_transcription(text)
        description = f'.{type} subtitles'
        log_message = f'Saved .{type} subtitles →'
        self._save_file(
            content=corrected_text,
            filepath=filepath,
            description=description,
            log_message=log_message
        )

    def save_rttm_annotations(
        self, 
        rttm: str, 
        filepath: Path
        ) -> None:
        self._save_file(
            content=rttm,
            filepath=filepath,
            description='.rttm annotations',
            log_message='Saved .rttm annotations →'
        )

    def save_results(
        self,
        result: dict,
        export_formats: List[str]
    ) -> List[Path]:
        """
        Write various output formats to disk based on the specified export formats.
        """
        output_filepath = Path(result['output_filepath'])
        written_filepaths = []
        
        # Apply corrections if they are provided
        if self.corrections and (
            self.corrections.simple or self.corrections.patterns
            ):
            for language, transcription in result.get('transcription', {}).items():
                # Correct the main transcription text
                original_text = transcription.get('text', '').strip()
                corrected_text = self.correct_transcription(original_text)
                result['transcription'][language]['text'] = corrected_text
                
                # Correct chunks and word dicts
                chunks = transcription.get('chunks', '')
                for c in chunks:
                    # Text chunk
                    c['text'] = self.correct_transcription(c['text'])
                    # Words
                    words = c.get('words', '')
                    for w in words:
                        w['word'] = self.correct_transcription(w['word'])

                # Correct speaker annotations if present
                if 'text_with_speaker_annotation' in transcription:
                    original_annotated = transcription['text_with_speaker_annotation']
                    corrected_annotated = self.correct_transcription(original_annotated)
                    result['transcription'][language]['text_with_speaker_annotation'] = corrected_annotated

        # Now, transcription_items reflects the corrected transcriptions
        transcription_items = result.get('transcription', {}).items()
    
        # Write .txt
        if 'txt' in export_formats:
            for language, transcription in transcription_items:
                fout = output_filepath.parent / f"{output_filepath.name}_{language}.txt"
                self.save_txt(
                    transcription,
                    filepath=fout
                )
                written_filepaths.append(str(fout))

        # Write subtitles (.srt, .vtt and .webvtt)
        subtitle_formats = {'srt', 'vtt', 'webvtt'}
        if subtitle_formats.intersection(export_formats):
            for language, transcription in transcription_items:
                # .srt subtitles
                if 'srt' in export_formats:
                    fout = output_filepath.parent / f"{output_filepath.name}_{language}.srt"
                    srt_text = create_subtitles(
                        transcription, 
                        type='srt'
                        )
                    self.save_subtitles(srt_text, type='srt', filepath=fout)
                    written_filepaths.append(str(fout))

                # .vtt / .webvtt subtitles
                if 'vtt' in export_formats or 'webvtt' in export_formats:
                    for subtitle_type in ['webvtt', 'vtt']:
                        fout = output_filepath.parent / f"{output_filepath.name}_{language}.{subtitle_type}"
                        vtt_text = create_subtitles(
                            transcription, 
                            type=subtitle_type, 
                            result=result
                            )
                        self.save_subtitles(
                            vtt_text, 
                            type=subtitle_type, 
                            filepath=fout
                            )
                        written_filepaths.append(str(fout))

        # Write annotated .txt with speaker annotations
        has_speaker_annotation = any(
            'text_with_speaker_annotation' in transcription
            for transcription in result['transcription'].values()
        )

        if 'txt' in export_formats and has_speaker_annotation:
            for language, transcription in transcription_items:
                if 'text_with_speaker_annotation' in transcription:
                    fout = output_filepath.parent / f"{output_filepath.name}_{language}_annotated.txt"
                    self.save_txt_with_speaker_annotation(
                        annotated_text=transcription['text_with_speaker_annotation'],
                        filepath=fout
                    )
                    written_filepaths.append(str(fout))

        # Write .rttm
        if 'rttm' in export_formats:
            # Create .rttm annotations
            rttm_dict = dict_to_rttm(result)

            for language, rttm_annotation in rttm_dict.items():
                fout = output_filepath.parent / f"{output_filepath.name}_{language}.rttm"
                self.save_rttm_annotations(
                    rttm=rttm_annotation,
                    filepath=fout
                    )
                written_filepaths.append(str(fout))

        # Write .json
        if 'json' in export_formats:
            fout = output_filepath.with_suffix('.json')
            written_filepaths.append(str(fout))
            result['written_files'] = written_filepaths
            self.save_json(result, filepath=fout)

        return written_filepaths


def create_subtitles(
    transcription_dict: dict, 
    type: str = 'srt', 
    result: dict = None
    ) -> str:
    """
    Converts a transcription dictionary into subtitle format (.srt or .webvtt).

    Args:
        transcription_dict (dict): Dictionary containing transcription data 
            with 'chunks'.
        sub_length (int, optional): Maximum duration in seconds for each 
            subtitle block.
        type (str, optional): Subtitle format, either 'srt' or 'webvtt'. 
            Default is 'srt'.

    Returns:
        str: Formatted subtitle text in the specified format.
    """
    subtitle_text = ''
    seg_id = 0
    
    for chunk in transcription_dict['chunks']:
        start_time = chunk['timestamp'][0]
        end_time = chunk['timestamp'][1]
        text = chunk['text'].replace('’', '\'')
        
        # Create .srt subtitles
        if type == 'srt':
            start_time_str = little_helper.format_time(
                start_time, 
                delimiter=','
                )
            end_time_str = little_helper.format_time(
                end_time, 
                delimiter=','
                )
            seg_id += 1
            subtitle_text += f"""{seg_id}\n{start_time_str} --> {end_time_str}\n{text.strip()}\n\n"""
        
        # Create .webvtt subtitles    
        elif type in ['webvtt', 'vtt']:
            start_time_str = little_helper.format_time(
                start_time, 
                delimiter='.'
                )
            end_time_str = little_helper.format_time(
                end_time, 
                delimiter='.'
                )

            if seg_id == 0:
                subtitle_text += f"WEBVTT {Path(result['output_filepath']).stem}\n\n"
                
                if type == 'vtt':
                    subtitle_text += 'NOTE transcribed with whisply\n\n'
                    subtitle_text += f"NOTE media: {Path(result['input_filepath']).absolute()}\n\n"
                
            seg_id += 1
            subtitle_text += f"""{seg_id}\n{start_time_str} --> {end_time_str}\n{text.strip()}\n\n"""
        
    return subtitle_text

def dict_to_rttm(result: dict) -> dict:
    """
    Converts a transcription dictionary to RTTM file format.
    """
    file_id = result.get('input_filepath', 'unknown_file')
    file_id = Path(file_id).stem
    rttm_dict = {}

    # Iterate over each available language
    for lang, transcription in result.get('transcription', {}).items():
        lines = []
        current_speaker = None
        speaker_start_time = None
        speaker_end_time = None

        chunks = transcription.get('chunks', [])

        # Collect all words from chunks
        all_words = []
        for chunk in chunks:
            words = chunk.get('words', [])
            all_words.extend(words)

        # Sort all words by their start time
        all_words.sort(key=lambda w: w.get('start', 0.0))

        for word_info in all_words:
            speaker = word_info.get('speaker', 'SPEAKER_00')
            word_start = word_info.get('start', 0.0)
            word_end = word_info.get('end', word_start)

            if speaker != current_speaker:
                # If there is a previous speaker segment, write it to the RTTM
                if current_speaker is not None:
                    duration = speaker_end_time - speaker_start_time
                    rttm_line = (
                        f"SPEAKER {file_id} 1 {speaker_start_time:.3f} {duration:.3f} "
                        f"<NA> <NA> {current_speaker} <NA>"
                    )
                    lines.append(rttm_line)

                # Start a new speaker segment
                current_speaker = speaker
                speaker_start_time = word_start
                speaker_end_time = word_end
            else:
                # Extend the current speaker segment
                speaker_end_time = max(speaker_end_time, word_end)

        # Write the last speaker segment to the RTTM
        if current_speaker is not None:
            duration = speaker_end_time - speaker_start_time
            rttm_line = (
                f"SPEAKER {file_id} 1 {speaker_start_time:.3f} {duration:.3f} "
                f"<NA> <NA> {current_speaker} <NA>"
            )
            lines.append(rttm_line)

        rttm_content = "\n".join(lines)
        rttm_dict[lang] = rttm_content

    return rttm_dict



from rich import print

WHISPER_MODELS = {
    # Models for faster-whisper / whisperx: https://huggingface.co/Systran
    # Models for insanely-fast-whisper: https://huggingface.co/openai
    'tiny': {
        'faster-whisper': 'tiny',
        'insane-whisper': 'openai/whisper-tiny',
        'whisperx': 'tiny'
        },
    'tine-en': {
        'faster-whisper': 'tiny.en',
        'insane-whisper': 'openai/whisper-tiny.en',
        'whisperx': 'tiny.en'
        },        
    'base': {
        'faster-whisper': 'base',
        'insane-whisper': 'openai/whisper-base',
        'whisperx': 'base'
        },
    'base-en': {
        'faster-whisper': 'base.en',
        'insane-whisper': 'openai/whisper-base.en',
        'whisperx': 'base.en'
        },
    'small': {
        'faster-whisper': 'small',
        'insane-whisper': 'openai/whisper-small',
        'whisperx': 'small'
        },
    'small-en': {
        'faster-whisper': 'small.en',
        'insane-whisper': 'openai/whisper-small.en',
        'whisperx': 'small.en'
        },
    'distil-small-en': {
        'faster-whisper': 'distil-small.en', 
        'insane-whisper': 'distil-whisper/distil-small.en',
        'whisperx': None
        },
    'medium': {
        'faster-whisper': 'medium', 
        'insane-whisper': 'openai/whisper-medium',
        'whisperx': 'medium'
        },
    'medium-en': {
        'faster-whisper': 'medium.en', 
        'insane-whisper': 'openai/whisper-medium.en',
        'whisperx': 'medium.en'
        },
    'distil-medium-en': {
        'faster-whisper': 'distil-medium.en', 
        'insane-whisper': 'distil-whisper/distil-medium.en',
        'whisperx': None
        },
    'large': {
        'faster-whisper': 'large', 
        'insane-whisper': 'openai/whisper-large',
        'whisperx': 'large'
        },
    'large-v1': {
        'faster-whisper': 'large-v1', 
        'insane-whisper': 'openai/whisper-large-v1',
        'whisperx': 'large-v1'
        },
    'large-v2': {
        'faster-whisper': 'large-v2', 
        'insane-whisper': 'openai/whisper-large-v2',
        'whisperx': 'large-v2'
        },
    'distil-large-v2': {
        'faster-whisper': 'distil-large-v2', 
        'insane-whisper': 'distil-whisper/distil-large-v2',
        'whisperx': None
        },
    'large-v3': {
        'faster-whisper': 'large-v3', 
        'insane-whisper': 'openai/whisper-large-v3',
        'whisperx': 'large-v3'
        },
    'distil-large-v3': {
        'faster-whisper': 'distil-large-v3', 
        'insane-whisper': 'distil-whisper/distil-large-v3',
        'whisperx': None
        },
    'large-v3-turbo': {
        'faster-whisper': 'deepdml/faster-whisper-large-v3-turbo-ct2', 
        'insane-whisper': 'openai/whisper-large-v3-turbo',
        'whisperx': None
        },
}

def ensure_model(model: str) -> bool:
    return model in WHISPER_MODELS

def is_model_supported(model: str, implementation: str) -> bool:
    return True if WHISPER_MODELS.get(model)[implementation] != None else False

def set_supported_model(model: str, implementation: str) -> str:
    if not is_model_supported(model, implementation):
        print(f'[deep_pink3]→ Model "{model}" is not available for this task / implementation → Using default model "large-v2".')
        return WHISPER_MODELS.get("large-v2")[implementation]
    return WHISPER_MODELS.get(model)[implementation]
